from easy_tg_bot import get_secret, default_roles


# Telegram IDs of (super)admins
# https://t.me/getmyid_bot
MY_ID = get_secret("MY_ID")

# Default roles
# Possible keys: stuff in you .env or env variable with the name 
# Possible values: superadmin, admin, user, banned
ROLES = {
    "MY_ID": "superadmin"
}

# Add default roles 
default_roles(ROLES)



from telegram import Update, InlineKeyboardButton, InlineKeyboardMarkup
from telegram.ext import (
    CallbackContext,
    MessageHandler,
    filters,
    ConversationHandler,
    CallbackQueryHandler,
)

from .roles import (
    DEFAULT_ADMIN_ROLES,
    check_role,
    add_role,
    is_bot_closed,
    close_bot_for_users,
    open_bot_for_users,
    get_people_layout,
)
from .admin import admin
from .text_handler import text_handler
from .utils.utils import get_keyboard, get_info_from_query
from .send import send_message
from .send_with_navigation import send_page_with_navigation
from .mylogging import logger
from .decorators import register_conversation_handler, button_callback


class ManageRoleConverstion:
    def __init__(self, prefix, role, main_func=lambda context, user_id: None):
        self.prefix = prefix
        self.role = role
        self.main_func = main_func
        self.restrict_function = lambda u, c: check_role(c, DEFAULT_ADMIN_ROLES)
        self.ADD_ID = 0
        self.entry_point = f"{prefix}_{role}"
        self.bot_handler = ConversationHandler(
            entry_points=[
                CallbackQueryHandler(
                    self.start_conversation, pattern=rf"^{self.entry_point}$"
                )
            ],
            states={
                self.ADD_ID: [
                    MessageHandler(filters.TEXT, self.handle_message),
                ]
            },
            fallbacks=[
                CallbackQueryHandler(self.cancel, pattern=r"^cancel$"),
            ],
            name=self.entry_point,
            persistent=True,
        )

    async def cancel(self, update: Update, context: CallbackContext):
        # Answer callback
        if update:
            query = update.callback_query
            if query:
                await query.answer()

        # Check role
        if not self.restrict_function(update, context):
            return ConversationHandler.END

        await admin_manage_users_button(update, context)
        return ConversationHandler.END

    async def start_conversation(self, update: Update, context: CallbackContext):
        # Answer callback
        if update:
            query = update.callback_query
            if query:
                await query.answer()

        # Check role
        if not self.restrict_function(update, context):
            return ConversationHandler.END

        keyboard = get_keyboard(context, back_button_callback="cancel")
        await send_message(
            update,
            context,
            keyboard=keyboard,
            text_string_index=f"{self.entry_point}_text",
        )

        return self.ADD_ID

    async def handle_message(self, update: Update, context: CallbackContext) -> int:
        if not self.restrict_function(update, context):
            return ConversationHandler.END

        message = update.message.text
        ids = [i.strip() for i in message.split(",")]

        # validate id
        try:
            validation_result = [int(user_id) for user_id in ids]
        except Exception:
            await send_message(update, context, text_string_index="id_must_be_a_number", replace=False)
            return await self.start_conversation(update, context)

        results = []
        for user_id in ids:
            # can't change superadmins role
            if check_role(context, ("superadmin",), user_id):
                pass

            else:
                try:
                    result = self.main_func(context, user_id)
                    results.append(result)
                    logger.info(f"Success {self.entry_point}")
                except Exception as e:
                    results.append(None)
                    logger.error(f"Fail {self.entry_point}")

        if not all(results):
            await send_message(
                update, context, text_string_index=f"{self.entry_point}_fail", replace=False
            )
            return await self.start_conversation(update, context)
        else:
            await send_message(
                update,
                context,
                text_string_index=f"{self.entry_point}_success",
                replace=False,
            )
            return await self.cancel(update, context)


admin_convo_handlers = [
    ManageRoleConverstion("add", "admin", lambda c, u_id: add_role(c, "admin", u_id)),
    ManageRoleConverstion("add", "user", lambda c, u_id: add_role(c, "user", u_id)),
    ManageRoleConverstion("del", "user", lambda c, u_id: add_role(c, "banned", u_id)),
]
for h in admin_convo_handlers:
    register_conversation_handler(h.bot_handler)


@button_callback(allowed_roles=DEFAULT_ADMIN_ROLES)
async def admin_manage_users_button(update, context):
    # menu
    buttons = []

    # Open/close bot
    if is_bot_closed(context):
        b_text = text_handler.get_text(context, "open_bot")
    else:
        b_text = text_handler.get_text(context, "close_bot")

    buttons += [
        [
            InlineKeyboardButton(
                b_text,
                callback_data="open_close_bot",
            )
        ]
    ]

    # Get users
    buttons += [
        [
            InlineKeyboardButton(
                text_handler.get_text(context, "get_users_by_role_menu"),
                callback_data="get_users_by_role_menu",
            )
        ]
    ]

    # crud users
    options = {
        text_handler.get_text(context, h.entry_point): h.entry_point
        for h in admin_convo_handlers
    }
    buttons += [
        [
            InlineKeyboardButton(
                option_name,
                callback_data=option,
            )
        ]
        for option_name, option in options.items()
    ]

    # TODO forget users

    # back
    buttons += [
        [
            InlineKeyboardButton(
                text_handler.get_text(context, "back_button"),
                callback_data="admin_cancel",
            )
        ]
    ]

    keyboard = InlineKeyboardMarkup(buttons)
    return await send_message(
        update,
        context,
        keyboard=keyboard,
        text_string_index="admin_manage_users_header",
    )


@button_callback(allowed_roles=DEFAULT_ADMIN_ROLES)
async def admin_cancel(update, context):
    return await admin(update, context)


@button_callback(allowed_roles=DEFAULT_ADMIN_ROLES)
async def open_close_bot(update, context):
    if is_bot_closed(context):
        open_bot_for_users(context)
        await send_message(update, context, text_string_index="bot_opened", replace=False)
        logger.info("Bot has been opened")
    else:
        close_bot_for_users(context)
        await send_message(update, context, text_string_index="bot_closed", replace=False)
        logger.info("Bot has been closed")

    return await admin_manage_users_button(update, context)


@button_callback(allowed_roles=DEFAULT_ADMIN_ROLES)
async def get_users_by_role_menu(update, context):
    prefix = "admin_get_users"
    options = ("admin", "user", "banned")
    buttons = []

    # List options
    buttons += [
        [
            InlineKeyboardButton(
                text_handler.get_text(context, f"{prefix}_{option}"),
                callback_data=f"{prefix}_{option}",
            )
        ]
        for option in options
    ]

    # Go back
    buttons += [
        [
            InlineKeyboardButton(
                text_handler.get_text(
                    context,
                    text_string_index="back_button",
                ),
                callback_data="admin_manage_users_button",
            )
        ]
    ]
    return await send_message(
        update,
        context,
        keyboard=InlineKeyboardMarkup(buttons),
        text_string_index="admin_get_users_header",
    )


@button_callback(allowed_roles=DEFAULT_ADMIN_ROLES)
async def admin_get_users(update, context):
    role = await get_info_from_query(update, "admin_get_users")
    return await send_page_with_navigation(
        update,
        context,
        lines_header=f"admin_get_users_header_{role}",
        get_lines_func=lambda context: get_people_layout(context, role),
        lines_to_text_func=lambda l: "\n\n".join(l),
        callback_func=get_users_by_role_menu,
        clear_cache=True,
    )



# Server utils
import os
import click
from .settings import get_secret


@click.group()
@click.version_option()
def srv():
    """The server utils CLI tool."""


def files_from_server_to_my_windows(remote_path, local_path, user, ip, debug=False):
    """requires ssh key to be set up"""
    command = f"scp -r  {user}@{ip}:{remote_path} {local_path}"
    if debug:
        print(command)
    os.system(command)


def files_from_my_windows_to_server(remote_path, local_path, user, ip, debug=False):
    """requires ssh key to be set up"""
    command = f"scp -r {local_path} {user}@{ip}:{remote_path}"
    if debug:
        print(command)
    os.system(command)


@click.option("--debug", is_flag=True, help="Print the command to the console")
@srv.command()
def bring_files_from_server(debug):
    """requires ssh key to be set up"""
    local_path = get_secret("LOCAL_PATH")
    remote_path = get_secret("REMOTE_PATH")
    user = get_secret("SERVER_USER")
    ip = get_secret("SERVER_IP")
    return files_from_server_to_my_windows(remote_path, local_path, user, ip, debug)


@click.option("--debug", is_flag=True, help="Print the command to the console")
@srv.command()
def bring_files_to_server(debug):
    """requires ssh key to be set up"""
    local_path = get_secret("LOCAL_PATH")
    remote_path = get_secret("REMOTE_PATH")
    user = get_secret("SERVER_USER")
    ip = get_secret("SERVER_IP")
    return files_from_my_windows_to_server(remote_path, local_path, user, ip, debug)



from telegram import Update
from telegram.ext import (
    CallbackContext,
    CallbackQueryHandler,
    ConversationHandler,
    MessageHandler,
    filters,
)

from .put_file_conv import PutFileConversation
from .mylogging import logger
from .utils.context_logic import put_chat_data, get_chat_data
from .utils.utils import get_keyboard, get_info_from_query, end_conversation
from .send import send_message
from .text_handler import text_handler
from .validate_text_file import DEFAULT_NOT_FOUND_TEXT

from .roles import check_role, DEFAULT_ALLOWED_ROLES
# entry point for the conversations
from .admin import ADD_INTRO_VID_BUTTON
from .video_handler import intro_video_handler

# annoying warning
from warnings import filterwarnings
from telegram.warnings import PTBUserWarning

filterwarnings(
    action="ignore", message=r".*CallbackQueryHandler", category=PTBUserWarning
)


class PutVideoConversation(PutFileConversation):
    def __init__(self, file_handler, entry_point_query_pattern):
        super().__init__(file_handler, entry_point_query_pattern)

        self.put_file_message = f"put_{file_handler.prefix}_video_message"  # text.xlsx
        self.error_message = "error_processing_video"  # text.xlsx
        self.success_message = "success_adding_video"  # text.xlsx

        self.INPUT_LAN = 1
        self.lan_prefix = f"{file_handler.prefix}_vlan"
        self.cached_lan_key = f"current_{self.lan_prefix}"
        self.video_path_dict_key = f"{file_handler.prefix}_video_dict"
        self.video_id_dict_key = f"{file_handler.prefix}_video_id_dict"
        self.bot_handler = ConversationHandler(
            entry_points=[
                CallbackQueryHandler(
                    self.start_conversation,
                    pattern=f"^{self.entry_point_query_pattern}$",
                )
            ],
            states={
                self.INPUT_LAN: [
                    CallbackQueryHandler(
                        self.language_choice, pattern=rf"^{self.lan_prefix}_"
                    )
                ],
                self.INPUT_FILE: [MessageHandler(filters.VIDEO, self.receive_file)],
            },
            fallbacks=[CallbackQueryHandler(self.end, pattern=r"^end$")],
            name=f"put_{self.file_handler.file_key}",
            persistent=True,
        )

    async def start_conversation(self, update: Update, context: CallbackContext):
        # Answer callback
        if update:
            query = update.callback_query
            if query:
                await query.answer()

        # Check role
        if not self.restrict_function(update, context):
            return ConversationHandler.END  # TODO decorators

        languages = text_handler.get_languages(context)
        keyboard = get_keyboard(
            context,
            options=languages,
            prefix=self.lan_prefix,
            back_button_callback="end",
        )
        await send_message(update, context, keyboard=keyboard, text_string_index="intro_video_language_choice") # text.xlsx
        return self.INPUT_LAN

    async def language_choice(self, update: Update, context: CallbackContext):
        # Answer callback
        if update:
            query = update.callback_query
            if query:
                await query.answer()

        # Check role
        if not self.restrict_function(update, context):
            return ConversationHandler.END 

        info = await get_info_from_query(update, self.lan_prefix)
        put_chat_data(context, self.cached_lan_key, info)
        keyboard = get_keyboard(context, back_button_callback="end") 
        await send_message(update, context, keyboard=keyboard, text_string_index=self.put_file_message)
        return self.INPUT_FILE

    async def receive_file(self, update: Update, context: CallbackContext):
        if not self.restrict_function(update, context):
            return ConversationHandler.END  

        # This goes after the language choice
        chat_data = get_chat_data(context)
        current_lan = chat_data.get(self.cached_lan_key)
        if not current_lan:
            logger.error("Error regisreting the video: language missing")
            await end_conversation(
                update,
                context,
                last_message_text_string_index=self.error_message,
                clean_up=True,
            )
            return await self.end(update, context)

        try:
            file_path = await self.file_handler.download_file(update)
            # Save intro video
            iterabl = self.get_lan_to_video_path_dict(context)
            old_file_path = iterabl.pop(current_lan, None)
            if old_file_path:
                await self.file_handler.delete_file(old_file_path)
                id_iterabl = self.get_lan_to_video_id_dict(context)
                id_iterabl.pop(current_lan, None)

            iterabl[current_lan] = file_path
            context.bot_data[self.video_path_dict_key] = iterabl

            await end_conversation(
                update,
                context,
                last_message_text_string_index=self.success_message,
                clean_up=True
            )
            await self.send_intro_video(update, context)
            logger.info("Intro video added")
            return await self.end(update, context)

        except Exception as e:
            logger.error(f"Error recieving the file: {str(e)}")
            return await self.file_error_response(
                update, context, file_path, self.error_message
            )

    def get_lan_to_video_path_dict(self, context):
        return context.bot_data.get(self.video_path_dict_key, {})

    def get_lan_to_video_id_dict(self, context):
        return context.bot_data.get(self.video_id_dict_key, {})

    def get_video_path(self, context):
        lan = text_handler.get_user_language(context)
        dic = self.get_lan_to_video_path_dict(context)
        return dic.get(lan)

    def get_video_id(self, context):
        lan = text_handler.get_user_language(context)
        dic = self.get_lan_to_video_id_dict(context)
        return dic.get(lan)

    async def send_intro_video(self, update, context):
        if not check_role(context, DEFAULT_ALLOWED_ROLES):
            return ConversationHandler.END 

        # If there is no caption, there is no intro video
        caption, parse_mode = text_handler.get_text_and_parse_mode(
            context, "intro_video_caption"
        )
        if caption == DEFAULT_NOT_FOUND_TEXT:
            # caption = None
            return

        vid_id = self.get_video_id(context)
        if vid_id:
            try:
                return await context.bot.send_video(
                    chat_id=update.effective_chat.id,
                    video=vid_id,
                    caption=caption,
                    parse_mode=parse_mode,
                )
            except Exception as e:
                logger.warning(f"Sending the intro video from file_id: {e}")

        vid_path = self.get_video_path(context)
        if vid_path:
            try:
                with open(vid_path, "rb") as video_file:
                    message = await context.bot.send_video(
                        chat_id=update.effective_chat.id,
                        video=video_file,
                        caption=caption,
                        parse_mode=parse_mode,
                    )
                # Store the file_id for future use
                if message and message.video:
                    id_iterabl = self.get_lan_to_video_id_dict(context)
                    id_iterabl[text_handler.get_user_language(context)] = (
                        message.video.file_id
                    )
                    context.bot_data[self.video_id_dict_key] = (
                        id_iterabl
                    )
                return message
            except Exception as e:
                logger.error(f"Error sending the intro video from file: {e}")
        return None


put_intro_video_file_conv = PutVideoConversation(intro_video_handler, ADD_INTRO_VID_BUTTON)
put_intro_video_file_conv.register_handler()



from telegram import Update, InlineKeyboardMarkup
from telegram.ext import CallbackContext

from .utils.context_logic import get_chat_data, put_chat_data
from .text_handler import text_handler
from .mylogging import logger, time_log_decorator


# TODO list of reserved keys for the bot
LAST_MESSAGE_ID = "lkmid"


# Caching logic
async def kb_cache_logic(
    update, context, sent_message_id, user_id=None
):
    if sent_message_id:
        # Mute a previous active window so only one active window can be in a chat
        await mute_last_active_msg(update, context, user_id)
        # Store the message ID of the new start menu message
        await register_last_active_msg(context, sent_message_id, user_id)


async def register_last_active_msg(context: CallbackContext, message_id, user_id=None):
    put_chat_data(context, LAST_MESSAGE_ID, message_id, user_id)


async def mute_last_active_msg(update: Update, context: CallbackContext, user_id=None):
    chat_data = get_chat_data(context, user_id)
    message_id = chat_data.get(LAST_MESSAGE_ID)
    chat_id = user_id or update.effective_chat.id
    if message_id:
        try:
            await context.bot.delete_message(
                chat_id=chat_id,
                message_id=message_id,
            )
            chat_data.pop(LAST_MESSAGE_ID)
        except Exception as e:
            logger.warning(f"Error muting last active keyboard {chat_id}: {str(e)}") # TODO remove user/chat_id from the logs


# Send raw
async def edit_message_no_cache(
    update: Update,
    context: CallbackContext,
    text=None,
    parse_mode=None,
    disable_web_page_preview=None,
    keyboard=None,
    user_id=None,
):
    if user_id is None:
        query = update.callback_query
        if query:
            return await query.message.edit_text(
            # doesn't fucking work:
            #return await context.bot.edit_message_text(
            #chat_id=chat_id,
            #message_id=message_id,
            reply_markup=keyboard,
            text=text,
            parse_mode=parse_mode,
            disable_web_page_preview=disable_web_page_preview,
        )
        raise ValueError("Query not found")
    else:
        raise ValueError("Can't edit query for a user")


async def send_message_no_cache(
    update: Update,
    context: CallbackContext,
    text,
    keyboard=None,
    user_id=None,
    **kwargs,
):
    chat_id = user_id or update.effective_chat.id
    try:
        sent_message = await context.bot.send_message(chat_id=chat_id, text=text, reply_markup=keyboard, **kwargs)
        return sent_message.message_id
    except Exception as e:
        logger.error(f"Error sending a keyboard - {chat_id}: {e}")
        return 0

async def send_message(
    update,
    context,
    text= None,
    parse_mode=None,
    text_string_index=None,
    keyboard = None,
    user_id=None,
    replace=True,
    new=False,
    disable_web_page_preview=True,
    **kwargs,
):
    """
    - either text, parse mode kwargs, or text_string_index,
    in the latter case it will take text, parse_mode from the text.xlsx;
    - user_id = None for sending within user context;
    - replace = False for ignoring caching last message id logic;
    - new = True means no editing, but sending using caching logic;
    """
    # TODO validate kwargs usage
    # Send
    if text_string_index:
        text, parse_mode = text_handler.get_text_and_parse_mode(
            context, text_string_index, user_id
        )

    if replace and not new:
        try: 
            return await edit_message_no_cache(
                update,
                context,
                text=text,
                keyboard=keyboard,
                parse_mode=parse_mode,
                disable_web_page_preview=disable_web_page_preview,
                user_id=user_id,
            )
        except Exception as e:
            # logger.warning(f"Error editing a message: {e}")
            pass # TODO

    sent_message_id = await send_message_no_cache(
        update,
        context,
        text,
        keyboard=keyboard,
        parse_mode=parse_mode,
        disable_web_page_preview=disable_web_page_preview,
        user_id=user_id,
        **kwargs,
    )

    if replace:
        await kb_cache_logic(update, context, sent_message_id, user_id)

    return sent_message_id


# Utils + old name convention
async def send_video_from_file_id_raw(
    update, context, keyboard, file_id, caption, parse_mode, user_id=None
):
    if user_id is None:
        user_id = update.effective_chat.id

    sent_message = await context.bot.send_video(
        chat_id=user_id,
        reply_markup=keyboard,
        video=file_id,
        caption=caption,
        parse_mode=parse_mode,
    )
    return sent_message.message_id


async def send_video_from_file_id(
    update,
    context,
    keyboard,
    file_id,
    new = True,
    user_id=None,
    caption=None,
    parse_mode=None,
):
    """Sends video as a keyboard"""
    # Send
    # TODO edit video
    sent_message_id = await send_video_from_file_id_raw(
        update, context, keyboard, file_id, caption, parse_mode
    )
    if new: 
        await kb_cache_logic(
            update, context, sent_message_id, user_id
        )


# Competability with other versions
# send_text, send_keyboard, send_text_raw, send_keyboard_raw - deprecated
async def send_keyboard_raw(
    update: Update,
    context: CallbackContext,
    keyboard: InlineKeyboardMarkup,
    text: str,
    clear_after=True,
    clear_before=True,
    user_id=None,
    parse_mode=None,
    disable_web_page_preview=True,
    **kwargs,
):
    if clear_before and clear_after:
        replace, new = True, False
    else:
        replace, new = False, False

    return await send_message(
        update,
        context,
        text=text,
        parse_mode=parse_mode,
        keyboard=keyboard,
        user_id=user_id,
        replace=replace,
        new=new,
        disable_web_page_preview=disable_web_page_preview,
        **kwargs,
    )


async def send_text_raw(
    update: Update, context: CallbackContext, text: str, user_id=None, **kwargs
):
    sent_message = await context.bot.send_message(
        chat_id=user_id or update.effective_chat.id, text=text, **kwargs
    )
    return sent_message.message_id


async def send_text(
    update: Update,
    context: CallbackContext,
    text_string_index: str,
    user_id=None,
    **kwargs,
):
    TEXT, PARSE_MODE = text_handler.get_text_and_parse_mode(
        context, text_string_index, user_id
    )
    return await send_text_raw(
        update,
        context,
        TEXT,
        user_id=user_id,
        parse_mode=PARSE_MODE,
        # disable_web_page_preview=True,
        **kwargs,
    )


async def send_keyboard(
    update: Update,
    context: CallbackContext,
    keyboard: InlineKeyboardMarkup,
    text_string_index: str,
    user_id=None,
    clear_after=True,
    clear_before=True,
    **kwargs,
):
    TEXT, PARSE_MODE = text_handler.get_text_and_parse_mode(
        context, text_string_index, user_id
    )
    return await send_keyboard_raw(
        update,
        context,
        keyboard,
        TEXT,
        user_id=user_id,
        clear_after=clear_after,
        clear_before=clear_before,
        parse_mode=PARSE_MODE,
        disable_web_page_preview=True,
        **kwargs,
    )



import asyncio
import json
import io
from random import randint

from .send import send_message
from .roles import get_all_people, get_people
from .decorators import command
from .mylogging import get_time
from . import app
from . import settings


# MENU
INFO_LINES = [
    " - /database",
    # " - /bot_restart", # TODO
    " - /bot_down",
    " - /start_notification",
]


# Admin utils
async def send_to_superadmin(update, context, info):
    for user_id in get_people(context, "superadmin"):
        await send_message(update, context, text=info, user_id=user_id, replace=False)


async def send_notification(update, context, notification_txt_id):
    all_users = get_all_people(context)
    await send_message(update, context, text="Mailing users...", replace=False)
    for user_id in all_users:
        await asyncio.sleep(randint(1, 3))
        await send_message(update, context, text_string_index=notification_txt_id, user_id=user_id, replace=False)
    await send_message(update, context, text="Mailing users done.", replace=False)


@command(allowed_roles=("superadmin",))
async def superadmin(u, c):
    lines = [
        "🍑  SUPERADMIN MENU",
    ]
    lines += reversed(INFO_LINES)
    return await send_message(u, c, text = "\n".join(lines), replace=False)


@command(allowed_roles=("superadmin",))
async def start_notification(update, context):
    text_id = "start_notification"
    return await send_notification(update, context, text_id)


@command(allowed_roles=("superadmin",))
async def bot_down(update, context):
    text_id = "stop_notification"
    await send_notification(update, context, text_id)
    return await context.application.stop()


# @command(allowed_roles=("superadmin",))
# async def bot_restart(update, context):
#     await send_text_raw(update, context, "Stopping...")
#     await context.application.stop()
#     await context.application.initialize()
#     await send_text_raw(update, context, "Starting...")
#     await context.application.start()
#     await send_text_raw(update, context, "Success.")


@command(allowed_roles=("superadmin",))
async def database(update, context):
    time = get_time(True)
    for name, data in zip(
        ["bot", "chat", "user"],
        [
            app.application.bot_data,
            app.application.chat_data,
            app.application.user_data,
        ],
    ):
        data = dict(data)
        data_buffer = io.BytesIO()
        data_buffer.write(json.dumps(data, indent=4, default=str).encode())
        data_buffer.seek(0)
        # Send the files via telegram
        doc_stem= f"{settings.BOT_NAME}_{name}_data_{time}.json"[:50]
        await context.bot.send_document(
            chat_id=update.effective_chat.id,
            document=data_buffer,
            filename=f"{doc_stem}.json",
        )
        data_buffer.close()



# Convention: user_id = None means the user modifies their data themselves,
# otherwise it requires user telegram id and happens from the outside (admin banning a user, etc)

from ..mylogging import logger


def get_user_data(context, user_id = None):
    if user_id is None:
        return context.user_data
    else:
        return context.application.user_data.setdefault(int(user_id), {})

def put_user_data(context, key, value, user_id = None):
    user_data = get_user_data(context, user_id)
    user_data[key] = value

def get_chat_data(context, chat_id = None):
    if chat_id is None:
        return context.chat_data
    else:
        return context.application.chat_data.setdefault(int(chat_id), {})

def put_chat_data(context, key, value, chat_id = None):
    chat_data = get_chat_data(context, chat_id)
    chat_data[key] = value

# Conversation util
def clear_cache(context):
    """
    Cached items to clean always have prefix "current" and are stored in chat_data
    """
    try:
        chat_data = get_chat_data(context)
        for k in list(chat_data.keys()):
            if k.startswith("current"):
                chat_data.pop(k)

    except Exception as e:
        logger.warning(f"Error in clear_cache: {str(e)}")



import pandas as pd
import os   

from .mylogging import logger


DEFAULT_TEXT_FILE = os.path.join("./data", "text.xlsx")
PARSE_MODE_COLUMN = "PARSE MODE"
INDEX_COLUMN = "TEXT INDEX"
VALIDATION_MESSAGE = "validate_text_{}"  # text.xlsx
DEFAULT_NOT_FOUND_TEXT = "404"
IGNORE_ABSENSE_TEXT_STRINGS = ["intro_video_caption"]

NECCESSARY_COLUMNS = [PARSE_MODE_COLUMN, INDEX_COLUMN]
NECCESSARY_SHEETS = []


def clear_table_element(element):
    element = element.strip() if isinstance(element, str) else element
    if not element:
        element = pd.NA
    return element


def validate_dataframes(file_path, sheets, columns):
    df_sheets = pd.read_excel(file_path, sheet_name=None, dtype=str)

    if not all(s in df_sheets for s in sheets):
        msg = "Sheets in the file are missing!"
        raise ValueError(msg)

    if not all(column in v.columns for column in columns for v in df_sheets.values()):
        msg = "Wrong columns in the file!"
        raise ValueError(msg)

    column_names = [df.columns.tolist() for df in df_sheets.values()]
    if not all(columns == column_names[0] for columns in column_names):
        msg = "Column names are different in different sheets."
        raise ValueError(msg)


# Get rid of spaces
def read_file(file_path) -> dict:
    # Initial validation
    validate_dataframes(
        file_path,
        NECCESSARY_SHEETS,
        NECCESSARY_COLUMNS,
    )

    df_sheets = pd.read_excel(file_path, sheet_name=None, dtype=str, index_col=0)

    # Clear strings
    for k, v in df_sheets.items():

        # Strip whitespace from all string data in the DataFrame
        v = v.map(clear_table_element)

        # Strip whitespace from column names
        v.columns = v.columns.map(clear_table_element)

        # Strip whitespace from index names
        v.index = v.index.map(clear_table_element)
        v.index_name = INDEX_COLUMN

        # Drop rows where the index is empty
        v = v.dropna(axis=0, how="all")

        df_sheets[k] = v

    return df_sheets


def get_text_dictionary(file_name=DEFAULT_TEXT_FILE):
    df = pd.concat(read_file(file_name).values(), axis=0)
    return {column: {index: value for index, value in df[column].dropna().items()} for column in df.columns if not df[column].dropna().empty}


def get_languages_from_dict(text_dictionary):
    return [lan for lan in text_dictionary.keys() if lan != PARSE_MODE_COLUMN]


def get_main_language_from_dict(text_dictionary):
    return get_languages_from_dict(text_dictionary)[0]


def get_parse_mode_from_dict(text_dictionary, text_string_index):
    # All the languages have the same parse mode
    return text_dictionary.get(PARSE_MODE_COLUMN,{}).get(text_string_index)


def get_text_from_dict(language, text_dictionary, text_string_index):
    result = text_dictionary.get(language, {}).get(text_string_index)
    if result is None:
        main_bot_language = get_main_language_from_dict(text_dictionary)
        result = text_dictionary.get(main_bot_language,{}).get(text_string_index)
        if result is None:
            if text_string_index not in IGNORE_ABSENSE_TEXT_STRINGS:
                logger.error(f"Key Error in file_dict: {text_string_index}")
            return DEFAULT_NOT_FOUND_TEXT
    return result


def get_text_and_parse_mode_from_dict(language, text_dictionary, text_string_index):
    return get_text_from_dict(
        language, text_dictionary, text_string_index
    ), get_parse_mode_from_dict(text_dictionary, text_string_index)


def validate_text_file(file_path) -> tuple:
    # Success
    code = 200
    try:
        TEXT_DICT = get_text_dictionary(file_path)
        LANS = get_languages_from_dict(TEXT_DICT)
        MAIN_LAN = get_main_language_from_dict(TEXT_DICT)
        # TODO parse mode and text example

    except Exception as e:
        # Validation error
        code = 400
        logger.error(str(e))

    finally:
        result = code, VALIDATION_MESSAGE.format(code)
        logger.info(result[1])
        return result


if __name__ == "__main__":
    # Test the file reading
    print(get_text_dictionary())#["English"])
    print(validate_text_file(DEFAULT_TEXT_FILE))



import logging
from logging import handlers

import asyncio
import sys
import os
import time
from datetime import datetime
import pytz

import re
import functools

from . import settings


# Initialize the data folder
root_dir = "./data"
if not os.path.exists(root_dir):
        os.makedirs(root_dir)

NAME = os.path.join(root_dir, f"{settings.BOT_NAME}.log")

os.environ["PYTHONIOENCODING"] = "utf-8"

# Time convention
def get_time(string=False):
    if settings.TIME_ZONE:
        time = datetime.now(tz=pytz.timezone(str(settings.TIME_ZONE)))
    else:
        time = datetime.now()
    if string:
        return time.strftime("%d_%m_%Y_%H_%M_%S_%f")
    return time


# --------------------------------------UTILS-----------------------------------------------
def redacted(error_msg: str) -> str:
    """this is a filter function for errors to cover unnecessary info"""
    error_msg = re.sub(r"api-\w{4}-\w{4}-\w{4}", "[REDACTED-API-KEY]", error_msg)
    error_msg = re.sub(
        r"\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,7}\b",
        "[REDACTED]",
        error_msg,
    )
    error_msg = re.sub(r"(https://api.telegram.org/).*", r"\1 [REDACTED]", error_msg)
    return error_msg


# -------------------------------------SET UP-----------------------------------------------
class CustomStreamHandler(logging.StreamHandler):
    def __init__(self, stream=sys.stdout):
        super().__init__(stream=stream)

    def emit(self, record):
        try:
            msg = self.format(record)
            # Encode and directly write bytes to the underlying buffer of the stream
            # if it's available (this bypasses the encoding issues of the stream's write method)
            if hasattr(self.stream, "buffer"):
                self.stream.buffer.write(
                    msg.encode("utf-8") + self.terminator.encode("utf-8")
                )
            else:
                # Fallback to writing as a string if buffer is not available
                self.stream.write(msg + self.terminator)
            self.flush()
        except Exception:
            self.handleError(record)


# Windows problem? delay = True
# https://stackoverflow.com/questions/22459850/permissionerror-when-using-python-3-3-4-and-rotatingfilehandler
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(levelname)s - %(message)s",
    handlers=[
        CustomStreamHandler(),
        # # automatic log rotation
        handlers.RotatingFileHandler(NAME, maxBytes=2 * 1024 * 10, backupCount=5, delay=True),
    ],
)

# get rid of the annoying logs
logging.getLogger("http").propagate = False
logging.getLogger("httpx").propagate = False
logging.getLogger("aiohttp").propagate = False

logger = logging.getLogger(__name__)


# -----------------------------------UNIVERSAL DEBUG DECORATOR--------------------------------
def time_log_decorator(func):
    if asyncio.iscoroutinefunction(func):

        @functools.wraps(func)
        async def async_wrapper(*args, **kwargs):
            start_time = time.time()

            logger.info(f"Entering {func.__name__}")

            try:
                result = await func(*args, **kwargs)
            except Exception as e:
                logger.error(f"Error in {func.__name__}: {redacted(str(e))}")
                raise e

            else:
                elapsed_time = time.time() - start_time
                logger.info(
                    f"Exiting {func.__name__}. Execution time: {elapsed_time:.4f} seconds"
                )
                return result

        return async_wrapper
    else:

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            start_time = time.time()

            logger.info(f"Entering {func.__name__}")

            try:
                result = func(*args, **kwargs)
            except Exception as e:
                logger.error(f"Error in {func.__name__}: {redacted(str(e))}")
                raise e
            else:
                elapsed_time = time.time() - start_time
                logger.info(
                    f"Exiting {func.__name__}. Execution time: {elapsed_time:.4f} seconds"
                )
                return result

        return wrapper


# -----------------------------------------TEST DECORATOR-----------------------------------
@time_log_decorator
def test_function():
    print("Inside a synchronous function")


@time_log_decorator
async def async_test_function():
    print("Entering an asynchronous function")
    await asyncio.sleep(1)
    print("Exiting an asynchronous function")


# Test the decorator
if __name__ == "__main__":
    test_function()
    asyncio.run(async_test_function())



from telegram import (
    Update,
    KeyboardButton,
    ReplyKeyboardMarkup,
    ReplyKeyboardRemove,
)
from telegram.ext import (
    CallbackContext,
    CommandHandler,
    ConversationHandler,
    MessageHandler,
    CallbackQueryHandler,
    filters,
)
from asyncio import sleep

from .roles import start_permission, role_required, check_role, DEFAULT_ADMIN_ROLES
from .text_handler import text_handler
from .send import send_message
from .utils.utils import (
    get_info_from_query,
    get_keyboard,
    put_info_to_user_data,
    is_info_in_user_data
)

from .put_intro_video_conv import put_intro_video_file_conv

from .decorators import register_conversation_handler
from .mylogging import logger

# remove the annoying warning
from warnings import filterwarnings
from telegram.warnings import PTBUserWarning

filterwarnings(
    action="ignore", message=r".*CallbackQueryHandler", category=PTBUserWarning
)


# Constants
START_DONE_CALLBACK = None  # Put your start function here; preferably it involves send_keyboard
LANGUAGE_CHOICE, DATA_CONSENT = range(2)


async def start(update: Update, context: CallbackContext):
    if not start_permission(update, context):  # TODO allowed users
        return ConversationHandler.END

    # refresh keyboards
    await send_message(
        update,
        context,
        text_string_index="welcome_text",
        keyboard=ReplyKeyboardRemove(),
        replace=False
    )  
    # just in case
    return await send_lan_choice_keyboard(update, context)


@role_required()
async def send_lan_choice_keyboard(update: Update, context: CallbackContext):
    languages = text_handler.get_languages(context)
    keyboard = get_keyboard(context, options=languages, prefix="lan")
    await send_message(update, context, keyboard=keyboard, text_string_index="language_choice", new=True) 
    return LANGUAGE_CHOICE


@role_required()
async def language_choice(update: Update, context: CallbackContext):
    # Answer callback
    if update:
        query = update.callback_query
        if query:
            await query.answer()

    lan = await get_info_from_query(update, "lan")
    text_handler.assign_language_to_user(context, lan)
    if is_info_in_user_data(update, context) or check_role(context, DEFAULT_ADMIN_ROLES):
        return await end(update, context)
    return await send_data_consent_keyboard(update, context)


@role_required()
async def send_data_consent_keyboard(update: Update, context: CallbackContext):
    CONFIRM_BUTTON = text_handler.get_text(context, "start_confirm_button")
    EXIT_BUTTON = text_handler.get_text(context, "back_button")
    buttons = [
        [KeyboardButton(CONFIRM_BUTTON, request_contact=True)],
        [KeyboardButton(EXIT_BUTTON)],
    ]
    keyboard = ReplyKeyboardMarkup(
        buttons, one_time_keyboard=True, resize_keyboard=True
    )
    await send_message(update, context, keyboard=keyboard, text_string_index="data_policy", replace=False)
    return DATA_CONSENT


@role_required()
async def data_consent(update: Update, context: CallbackContext):
    put_info_to_user_data(update, context)
    logger.info("New user!")
    return await end(update, context, intro_vid=True)


@role_required()
async def not_data_consent(update: Update, context: CallbackContext):
    CONFIRM_BUTTON = text_handler.get_text(context, "start_confirm_button")
    if update.message.text != CONFIRM_BUTTON:
        await send_message(
            update, context, text_string_index="not_confirm_text", keyboard=ReplyKeyboardRemove()
        )
    else:
        logger.error("Error in not_data_consent")
    return ConversationHandler.END


@role_required()
async def end(update: Update, context: CallbackContext, intro_vid=False):
    # remove keyboard
    await send_message(
        update, context, text_string_index="start_confirm_text", keyboard=ReplyKeyboardRemove()
    )

    if intro_vid:
        await put_intro_video_file_conv.send_intro_video(update, context)
        await sleep(2)

    if START_DONE_CALLBACK is not None:
        await START_DONE_CALLBACK(update, context)  # <- has to use
    
    return ConversationHandler.END


# objects
start_conv_handler = ConversationHandler(
    entry_points=[CommandHandler("start", start)],
    states={
        LANGUAGE_CHOICE: [
            CallbackQueryHandler(language_choice, pattern=r"^lan_.*"),
        ],
        DATA_CONSENT: [
            MessageHandler(filters.CONTACT, data_consent),
        ],
    },
    fallbacks=[
        MessageHandler(filters.ALL, not_data_consent),
    ],
    name = "start_command",
    persistent = True
)

register_conversation_handler(start_conv_handler)



from functools import wraps
from telegram.ext import ConversationHandler

from .utils.context_logic import put_user_data, get_user_data
from .utils.utils import get_user_data_essential
from .settings import get_default_role


DEFAULT_ALLOWED_ROLES = ("superadmin", "admin", "user")
DEFAULT_ADMIN_ROLES = ("superadmin", "admin")


# Basic
def add_role(context, role, user_id = None):
    put_user_data(context, "role", role, user_id)
    return get_role(context, user_id)

def get_role(context, user_id = None):
    user_data = get_user_data(context, user_id)
    return user_data.get("role")

def check_role(context, allowed_roles, user_id=None):
    role = get_role(context, user_id)
    if not role:
        return
    if role == "banned":
        return
    return role in allowed_roles

def role_required(allowed_roles = DEFAULT_ALLOWED_ROLES):
    """A role decorator for the bot handlers that take update, context.
    A user has to start with /start to get a role and thus access."""
    def decorator(func):

        @wraps(func)
        async def wrapper(update, context, *args, **kwargs):
            if not check_role(context, allowed_roles):
                return ConversationHandler.END
            return await func(update, context, *args, **kwargs)

        return wrapper
    return decorator

# /start
def add_a_role_on_start(update, context):
    user_id = update.effective_user.id
    role = get_default_role(user_id)
    add_role(context, role)
    return role

def start_permission(update, context):
    current_role = get_role(context)
    if current_role == "banned":
        return False
    if update.effective_user.is_bot:
        # TODO ban
        return False
    if not current_role:
        if is_bot_closed(context):
            return False
        role = add_a_role_on_start(update, context)
        if role == "banned":
            return False
    return True

# Open/Close bot; open by default
def is_bot_closed(context):
    return context.bot_data.get("is_closed")

def close_bot_for_users(context):
    context.bot_data["is_closed"] = True
    return is_bot_closed(context)

def open_bot_for_users(context):
    return context.bot_data.pop("is_closed", None)

# Info
def get_people(context, role) -> list:
    general_user_data = context.application.user_data
    return [
        str(user_id)
        for user_id in general_user_data
        if general_user_data.get(user_id, {}).get("role", "") == role
    ]

def get_people_role_group(context, role_group):
    general_user_data = context.application.user_data
    return [
        str(user_id)
        for user_id in general_user_data
        if general_user_data.get(user_id, {}).get("role", "") in role_group
    ]


def get_people_layout(context, role):
    people_ids = get_people(context, role)
    info_lines = [
        f"{i+1}. {user_id}\n{get_user_data_essential(context, user_id)}"
        for i, user_id in enumerate(people_ids)
    ]
    return info_lines

def get_all_people(context) -> list:
    general_user_data = context.application.user_data
    return [
        str(user_id)
        for user_id in general_user_data
        if general_user_data.get(user_id, {}).get("role", "") != "banned"
    ]



# Stuff to use
__version__ = "0.1.0"
from .app import telegram_bot_polling
from .mylogging import time_log_decorator, logger, get_time
from .decorators import command, button_callback, register_conversation_handler
from .admin import admin, ADMIN_MENU
from .text_handler import text_handler
from .send import send_text, send_keyboard, send_text_raw, send_keyboard_raw
from .utils.utils import (
    get_keyboard,
    get_info_from_query,
    end_conversation,
)
from .validate_text_file import clear_table_element
from .file_handler import FileHandler
from .video_handler import VideoFileHandler
# for adding the handlers on init automatically
from .start import start_conv_handler
from .put_file_conv import PutFileConversation
from .put_intro_video_conv import PutVideoConversation
__version__ = "0.1.1"
from .utils.context_logic import (
    get_chat_data,
    put_chat_data,
    get_user_data,
    put_user_data,
    clear_cache,
)
from .send import send_video_from_file_id
from .roles import (
    role_required,
    check_role,
    DEFAULT_ADMIN_ROLES,
    DEFAULT_ALLOWED_ROLES,
    get_people_layout,
    get_people_role_group,
)
from .send_with_navigation import send_page_with_navigation, send_page_nav
# for adding the handlers on init automatically
from .manage_role_conv import ManageRoleConverstion
# Serverless lambda
from .app import lambda_handler
__version__ = "0.1.3"
from .decorators import message_handler
__version__ = "0.1.4"
from .app import application
from .settings import get_secret, default_roles
__version__ = "0.1.5"
# for adding the handlers on init automatically
from .mail_users_conv import message_mailing_conv_handler
from . import superadmin  # INFO_LINES for superadmin menu
from . import error
__version__ = "0.1.9"
from .send import send_message
__version__ = "0.2.1" 
# New command in cli for deployment in vultr! easy-tg-bot vultr
# requires a vultr server with docker installed and vultr container registry
__version__ = "0.2.7"
# get_secret_by_name is deprecated
# set-webhook command is deprecated
# get-webhook-info command is added
# generate-secret-token command is added
# variables renamed!
from .app import run_webhook_uvicorn, run_webhook_gunicorn
__version__ = "0.2.8"
# srv bring-files-to-server and srv bring-files-from-server commands added
# srv bring-files-to-server --debug
# TODO
# - limiter;
# - Forget user /forget_me; Delete user;
# - Hung convo bug - test
# - Connection error handler
# - Limiter
# - Replace 'nan' with None
# - Small Bug fix: Error in not_data_consent ?
# - Bug fix: None.log occasionally on run
# - /restart_me
# - optional number share
# - cache errors; cache messages;



import hashlib
from datetime import timedelta
from .mylogging import logger, get_time
from .superadmin import send_to_superadmin
from . import app


# Whithin Telegram
async def error(update, context, cache={}, max_cache_size = 30, cooldown_minutes = 30) -> None:
    error_msg = "General error handler - "
    try:
        # No internet during polling or a webhook issue
        if not update:
            error_msg += f"no Update found"
        # For callback queries
        elif update.callback_query:
            error_msg += f"a callback error occurred: {context.error}"
            # await update.callback_query.answer(msg, show_alert=True)
        # For regular messages
        elif update.message:
            error_msg += f"a message error occurred: {context.error}"
            # await update.message.reply_text(msg)
        else:
            error_msg += f"Update\n'{update}'\n...caused error\n'{context.error}'"
    except Exception as e:
        error_msg += f"error in the bot: {e}"

    logger.error(error_msg)

    # Clear cache
    current_time = get_time()
    cooldown = timedelta(minutes=cooldown_minutes)
    cache = {
        key: timestamp
        for key, timestamp in cache.items()
        if current_time - timestamp <= cooldown
    }

    # If the cache becomes too large
    if len(cache) >= max_cache_size:
        logger.info(f"General error handler - the cache if full: {max_cache_size} errors")
        return

    error_hash = hashlib.md5(error_msg.encode()).hexdigest()
    if error_hash not in cache:
        cache[error_hash] = current_time  # Update the cache
        await maybe_notify_superadmin(update, context, error_msg)
        return
    else:
        return 

async def maybe_notify_superadmin(update, context, error_msg):
    try:
        await send_to_superadmin(update, context, error_msg[:2000])
    except Exception as e:
        logger.error(f"General error handler - error notifying superadmin about the error:\n{e}")
    return 

app.application.add_error_handler(error)



from telegram import Update, InlineKeyboardButton, InlineKeyboardMarkup
from telegram.ext import (
    CallbackContext,
    MessageHandler,
    filters,
    ConversationHandler,
    CallbackQueryHandler,
)
import asyncio
from random import randint

from .mylogging import logger
from .roles import DEFAULT_ADMIN_ROLES, role_required, get_all_people, get_people, get_people_role_group
from .decorators import register_conversation_handler
from .text_handler import text_handler
from .send import send_message
from .utils.context_logic import get_chat_data, put_chat_data
from .utils.utils import clear_cache, get_keyboard, get_info_from_query
from .admin import admin, MAILING_BUTTON


# remove the annoying warning
from warnings import filterwarnings
from telegram.warnings import PTBUserWarning

filterwarnings(
    action="ignore", message=r".*CallbackQueryHandler", category=PTBUserWarning
)


# Constants
INPUT_MESSAGE, DELIVERY_OPTIONS, INPUT_IDS, CONFIRM = range(4)
MESSAGE_KEY = "current_message_func" # convention for clearing up


@role_required(DEFAULT_ADMIN_ROLES)
async def cancel(update: Update, context: CallbackContext):
    clear_cache(context)
    await admin(update, context)
    return ConversationHandler.END


@role_required(DEFAULT_ADMIN_ROLES)
async def start_conversation(update: Update, context: CallbackContext) -> int:
    query = update.callback_query
    await query.answer()
    keyboard = get_keyboard(context, back_button_callback="mailing_cancel")
    await send_message(update, context, keyboard=keyboard, text_string_index="mailing_input_message")
    return INPUT_MESSAGE


@role_required(DEFAULT_ADMIN_ROLES)
async def handle_message(update: Update, context: CallbackContext) -> int:
    message = update.message
    if message:
        put_chat_data(context, MESSAGE_KEY, message)
        return await send_delivery_options(update, context)
    logger.error("Error in handle_message in mailing.")
    return await cancel(update, context)


async def send_delivery_options(update: Update, context: CallbackContext):
    txt_ids_buttons = [
        "all_users",
        "only_users",
        "only_admins",
        "input_ids",
    ]
    buttons = [
        [
            InlineKeyboardButton(
                text_handler.get_text(context, txt_id),
                callback_data=f"mailing_deliv_choice_{txt_id}",
            )
        ] for txt_id in txt_ids_buttons
    ]
    buttons.append(
        [
            InlineKeyboardButton(
                text_handler.get_text(context, "back_button"),
                callback_data="mailing_cancel",
            )
        ]
    )
    keyboard = InlineKeyboardMarkup(buttons)
    await send_message(update, context, keyboard=keyboard, text_string_index="mailing_deliv_choice_message")
    return DELIVERY_OPTIONS


@role_required(DEFAULT_ADMIN_ROLES)
async def handle_delivery_options_choice(update: Update, context: CallbackContext):
    deliv_choice = await get_info_from_query(update, "mailing_deliv_choice")
    put_chat_data(context, "current_deliv_choice", deliv_choice)

    if deliv_choice == "input_ids":
        keyboard = get_keyboard(
            context, back_button_callback="mailing_cancel"
        )
        await send_message(
            update, context, keyboard=keyboard, text_string_index="mailing_input_ids_message"
        )
        return INPUT_IDS
    else:
        return await send_confirm(update, context)

async def send_confirm(update: Update, context: CallbackContext):
    txt_ids_buttons = ["mailing_confirm"]
    buttons = [
        [
            InlineKeyboardButton(
                text_handler.get_text(context, txt_id),
                callback_data=txt_id,
            )
        ]
        for txt_id in txt_ids_buttons
    ]
    buttons.append(
        [
            InlineKeyboardButton(
                text_handler.get_text(context, "back_button"),
                callback_data="mailing_cancel",
            )
        ]
    )
    keyboard = InlineKeyboardMarkup(buttons)
    await send_message(
        update, context, keyboard=keyboard, text_string_index="mailing_confirm_message"
    )
    return CONFIRM


@role_required(DEFAULT_ADMIN_ROLES)
async def handle_input_ids(update: Update, context: CallbackContext) -> int:
    text = update.message.text.strip()
    try:
        user_ids = [i.strip() for i in text.split(",")]
    except Exception as e:
        msg = f"Error handling ids: {e}"
        logger.error(msg)
        await send_message(update, context, text=msg, replace=False)
        return await cancel(update, context)
    put_chat_data(context, "current_user_ids", user_ids)
    return await send_confirm(update, context)


@role_required(DEFAULT_ADMIN_ROLES)
async def handle_confirm(update: Update, context: CallbackContext) -> int:
    query = update.callback_query
    await query.answer()
    if query.data.startswith("mailing_confirm"):
        chat_data = get_chat_data(context)
        current_deliv_choice = chat_data.pop("current_deliv_choice", "")
        GET_DELIV_FUNC = {
            "all_users": get_all_people,
            "only_users": lambda c: get_people(c, "user"),
            "only_admins": lambda c: get_people_role_group(c, DEFAULT_ADMIN_ROLES),
            "input_ids": lambda c: get_chat_data(c).pop("current_user_ids", []),
        }
        get_ids_func = GET_DELIV_FUNC.get(current_deliv_choice)
        if get_ids_func:
            return await send_message_to_users(update, context, get_ids_func)
        else:
            logger.error(f"Error in mailing: unable to find ids - {query.data}")

    logger.error(f"Error in mailing: confirm callback not found - {query.data}")
    return await cancel(update, context)


# Utils
async def send_message_to_users(update: Update, context: CallbackContext, get_current_user_group_ids_func):
    chat_data = get_chat_data(context)
    message = chat_data.pop(MESSAGE_KEY, None)
    if not message:
        logger.error("Error in mailing: message not found")
        return await cancel(update, context)

    bot = context.bot
    if message.photo:
        func = bot.send_photo
        kwargs = {
            "photo": message.photo[-1].file_id,
            "caption": message.caption
        }

    elif message.video:
        func = bot.send_video
        kwargs = {
            "video": message.video.file_id,
            "caption": message.caption
        }

    elif message.document:
        func = bot.send_document
        kwargs = {
            "document": message.document.file_id,
            "caption": message.caption
        }

    elif message.text:
        func = bot.send_message
        kwargs = {
            "text": message.text
        }

    else:
        logger.error(f"Error mailing message to all users: {message}")
        await send_message(update, context, text_string_index="mailing_wrong_format_message", replace=False)
        return INPUT_MESSAGE

    await send_something_to_users(update, context, get_current_user_group_ids_func, func, **kwargs)
    return await cancel(update, context)


async def send_something_to_users(update, context, get_current_user_group_ids_func, function, *args, **kwargs):
    await send_message(update, context, text_string_index="admin_mailing_in_progress")
    for user_id in get_current_user_group_ids_func(context):
        await asyncio.sleep(randint(1, 5))  # sleep to repsect api rates
        try:
            await function(*args, chat_id=int(user_id), **kwargs)
        except Exception as e:
            msg = f"{user_id}: {str(e)}"
            logger.error("Error in mailing - "+ msg)
            full_msg = (
                text_handler.get_text(context, "admin_mailing_error") + "\n" + msg
            )
            await send_message(update, context,text=full_msg, replace=False)
    return await send_message(update, context, text_string_index="admin_mailing_done")


# Object
message_mailing_conv_handler = ConversationHandler(
    entry_points=[
        CallbackQueryHandler(
            start_conversation, pattern=f"^{MAILING_BUTTON}$"
        )
    ],
    states={
        INPUT_MESSAGE: [
            MessageHandler(filters.ALL, handle_message),
        ],
        CONFIRM: [
            CallbackQueryHandler(handle_confirm, pattern=r"^mailing_confirm.*"),
        ],
        INPUT_IDS: [
            MessageHandler(filters.TEXT, handle_input_ids),
        ],
        DELIVERY_OPTIONS: [
            CallbackQueryHandler(
                handle_delivery_options_choice, pattern=r"^mailing_deliv_choice_.*"
            ),
        ],
    },
    fallbacks=[
        CallbackQueryHandler(cancel, pattern="^mailing_cancel$"),
    ],
    name="mailing",
    persistent=True,
)
register_conversation_handler(message_mailing_conv_handler)



from telegram import Update
from telegram.ext import CallbackContext
import aiofiles.os
import os

from .mylogging import logger, get_time


class FileHandler:
    def __init__(
        self,
        prefix,
        read_file_function=lambda file_path: {},
        validate_function=lambda file_path: (200, "validate_200"),     
    ):
        self.prefix = prefix
        self.default_path = f"{prefix}.xlsx"
        self.file_key = f"{prefix}_file"
        self.key_to_iter = f"{prefix}_iterable"
        # To call inside the class; to call outside, put extra arg in
        self.read_file = read_file_function # into something iterable
        self.validate_file = validate_function

    # Basics
    async def download_file(self, update: Update):
        file_extension = update.message.effective_attachment.file_name.split(".")[-1]
        file_name = self.prefix + "_from_" + get_time(string=True)
        file_path = f"{file_name}.{file_extension}"
        new_file = await update.message.effective_attachment.get_file()
        await new_file.download_to_drive("./data/" + file_path)
        return file_path

    async def delete_file(self, file_path):
        if file_path == self.default_path:
            return  # deleting of the example file does nothing
        try:
            path = os.path.join("./data", file_path)
            await aiofiles.os.remove(path)
            logger.info(f"Successfully deleted file: {file_path}")
        except OSError as e:
            logger.error(
                f"Error deleting file {file_path}: {e.strerror}"
            )

    # Send file
    async def send_file_general(
        self, update: Update, context: CallbackContext, file_path
    ):  
        try:
            path = os.path.join("./data", file_path)
            with open(path, "rb") as file:
                message = await context.bot.send_document(
                    chat_id=update.effective_chat.id,
                    document=file,
                    filename=file_path,
                )
            if update:
                query = update.callback_query
                if query:
                    await query.answer()

            return message.id if message else None

        except Exception as e:
            logger.error(f"Error sending a {self.prefix} file: {e}")
            return

    async def send_file(self, update: Update, context: CallbackContext):
        path = self.get_path(context)
        return await self.send_file_general(update, context, path)

    # Storing path of the file
    def get_path(self, context: CallbackContext) -> str:
        return context.bot_data.get(self.file_key, self.default_path)

    async def add_path(self, context: CallbackContext, new_path: str):
        """use only after validation"""
        path = self.get_path(context)
        await self.delete_file(path)
        context.bot_data[self.file_key] = new_path

    # Storing proccessed version of the file
    def put_file_iterable(self, context: CallbackContext):
        """use only after refreshing path"""
        path = self.get_path(context)
        path = os.path.join("./data", path)
        context.bot_data[self.key_to_iter] = self.read_file(path)

    def get_file_iterable(self, context: CallbackContext):
        if not context.bot_data.get(self.key_to_iter):
            self.put_file_iterable(context)
        return context.bot_data.get(self.key_to_iter)



import aiofiles
import os

from .mylogging import logger
from .file_handler import FileHandler


class VideoFileHandler(FileHandler):
    def __init__(self,prefix):
        self.prefix = prefix
        self.file_key = f"{prefix}_video"
        self.key_to_file_id = f"{prefix}_video_id"

    async def delete_file(self, file_path):
        if file_path is not None:
            try:
                path = os.path.join("./data", file_path)
                await aiofiles.os.remove(path)
                logger.info(f"Successfully deleted video: {file_path}")
            except OSError as e:
                logger.error(f"Error deleting video {file_path}: {e.strerror}")


intro_video_handler = VideoFileHandler("intro")



import os
import shutil


# Consts
DRAFTS_FOLDER = "templates"
DRAFTS_FOLDER_PATH = os.path.join(
    os.path.dirname(__file__), DRAFTS_FOLDER
)


def initialize_file_from_draft(file_name, root_dir):
    # Create files
    if not os.path.exists(root_dir):
        os.makedirs(root_dir)

    output_file = os.path.join(root_dir, file_name)
    if os.path.exists(output_file):
        return

    # Debug
    if not os.path.exists(DRAFTS_FOLDER_PATH):
        print(f"Templates folder '{DRAFTS_FOLDER_PATH}' does not exist.")
        return

    draft_path = os.path.join(DRAFTS_FOLDER_PATH, file_name)
    if not os.path.exists(draft_path):
        print(f"{draft_path} does not exist.")
        return 

    # Create files
    shutil.copy(draft_path, output_file)
    print(f"{file_name} file created.")

    return True


def create_vultr_deploy_yml_from_draft():
    github_actions_folder = ".github/workflows"
    github_actions_folder_path = os.path.join(
        os.getcwd() , github_actions_folder
    )
    if not os.path.exists(github_actions_folder_path):
        os.makedirs(github_actions_folder_path)

    deploy_yml_draft = "deploy.yml"
    deploy_yml_output_path = os.path.join(github_actions_folder_path, deploy_yml_draft)
    if os.path.exists(deploy_yml_output_path):
        return
    
    deploy_yml_path = os.path.join(DRAFTS_FOLDER_PATH, deploy_yml_draft)
    if not os.path.exists(deploy_yml_path):
        print(f"{deploy_yml_path} does not exist.")
        return

    shutil.copy(deploy_yml_path, deploy_yml_output_path)
    print(f"deploy.yml created in {github_actions_folder_path}.")



from telegram import InlineKeyboardButton, InlineKeyboardMarkup
import math

from .roles import role_required
from .decorators import button_callback
from .utils.context_logic import put_chat_data, get_chat_data
from .utils.utils import clear_cache
from .text_handler import text_handler
from .send import send_message

@role_required()
async def send_page_with_navigation(
    update,
    context,
    user_id=None,
    **kwargs,
):
    """
    GENERAL
    {
        "page": kwargs.get("start_page", 0),
        "per_page": kwargs.get("per_page", 5),
        "lines_header": kwargs.get("lines_header", ""),
        "get_lines_func": kwargs.get("get_lines_func", None),
        "callback_func": kwargs.get("callback_func", None),
        "clear_cache": kwargs.get("clear_cache", False)
    }

    TEXT
    {
        "lines_to_text_func": kwargs.get("lines_to_text_func"),
    }
    """
    prepare_navigation(context, **kwargs, user_id=user_id)
    return await send_page_nav(update, context, user_id)


def prepare_navigation(context, **kwargs):
    user_id = kwargs.pop("user_id", None)
    for key, value in kwargs.items():
        put_chat_data(context, f"current_{key}", value, user_id)


@role_required()
async def send_page_nav(update, context, user_id=None):
    """Triggers simple page refresh"""
    chat_data = get_chat_data(context, user_id)
    page = chat_data.get("current_page", 0)
    per_page = chat_data.get("current_per_page", 5)
    lines_header_index = chat_data.get("current_lines_header", "")
    lines = chat_data.get("current_get_lines_func", lambda c: [])(context)
    # TODO if not lines
    max_page = math.ceil(len(lines)/per_page)
    count = f"{page+1}/{max_page}" if max_page else "0/0"
    header = (
        text_handler.get_text(context, lines_header_index)
        + " "
        + count
    )

    start = page * per_page
    end = start + per_page

    small_list = [header] + lines[start:end]

    text_func = chat_data.get("current_lines_to_text_func")
    if text_func:
        text = text_func(small_list)
        buttons = get_navigation_buttons(context, page, per_page, lines)
        return await send_message(
            update, context, keyboard=InlineKeyboardMarkup(buttons), text=text, user_id=user_id
        )
    # TODO else


# Navigation buttons
def get_navigation_buttons(context, page, per_page, lines):
    amount_items = len(lines)

    # general
    buttons = []

    # navigation
    navigation_buttons = []

    if page >= 1:
        navigation_buttons.append(
            InlineKeyboardButton(
                "◀️",
                callback_data="previous_page",
            )
        )

    max_page = math.ceil(amount_items / per_page)
    if page + 1 < max_page:
        navigation_buttons.append(
            InlineKeyboardButton(
                "▶️",
                callback_data="next_page",
            )
        )

    if navigation_buttons:
        buttons.append(navigation_buttons)

    # back
    buttons.append(
        [
            InlineKeyboardButton(
                text_handler.get_text(context, "back_button"),
                callback_data="exit_navigation",
            )
        ]
    )
    return buttons


@button_callback()
async def next_page(update, context):
    chat_data = get_chat_data(context)
    page = chat_data.get("current_page", 0)
    put_chat_data(context, "current_page", page + 1)
    return await send_page_nav(update, context)


@button_callback()
async def previous_page(update, context):
    chat_data = get_chat_data(context)
    page = chat_data.get("current_page", 1)
    put_chat_data(context, "current_page", page - 1)
    return await send_page_nav(update, context)


@button_callback()
async def exit_navigation(update, context, user_id=None):
    chat_data = get_chat_data(context, user_id)
    callback_func = chat_data.pop("current_callback_func", None)
    clear_cache_flag = chat_data.pop("current_clear_cache", False)
    if clear_cache_flag:
        clear_cache(context)
    if callback_func:
        return await callback_func(update, context)
    # TODO else



from telegram import Update
from telegram.ext import (
    CallbackContext,
    CallbackQueryHandler,
    ConversationHandler,
    MessageHandler,
    filters,
)

import os

from .mylogging import logger
from .utils.utils import get_keyboard, end_conversation
from .send import send_message

from .file_handler import FileHandler
from .roles import check_role, DEFAULT_ADMIN_ROLES
# + entry point for the conversations
from .admin import admin, ADD_TEXT_BUTTON

# file handlers
from .text_handler import text_handler

# connect the handlers
from .decorators import register_conversation_handler

# annoying warning
from warnings import filterwarnings
from telegram.warnings import PTBUserWarning

filterwarnings(
    action="ignore", message=r".*CallbackQueryHandler", category=PTBUserWarning
)


class PutFileConversation:
    def __init__(
        self, file_handler: FileHandler, entry_point_query_pattern: str
    ):
        self.file_handler = file_handler
        self.entry_point_query_pattern = entry_point_query_pattern
        self.conversation_done_callback_function = admin
        self.restrict_function = lambda u, c: check_role(c, DEFAULT_ADMIN_ROLES)
        self.INPUT_FILE = 0

        self.put_file_message = f"put_{file_handler.prefix}_file_message"  # text.xlsx
        self.error_message = "error_processing_file"  # text.xlsx

        self.start_conversation_get_file_path = self.file_handler.get_path
        self.bot_handler = ConversationHandler(
            entry_points=[
                CallbackQueryHandler(
                    self.start_conversation,
                    pattern=f"^{self.entry_point_query_pattern}$",
                )
            ],
            states={
                self.INPUT_FILE: [
                    MessageHandler(
                        filters.Document.FileExtension("xlsx"), self.receive_file
                    )
                ],
            },
            fallbacks=[CallbackQueryHandler(self.end, pattern=r"^end$")],
            name=f"put_{self.file_handler.file_key}",
            persistent=True
        )

    async def end(self, update: Update, context: CallbackContext):
        # Answer callback
        if update:
            query = update.callback_query
            if query:
                await query.answer()

        # Check role
        if not self.restrict_function(update, context):
            return ConversationHandler.END  # TODO decorators

        await self.conversation_done_callback_function(update, context)
        return ConversationHandler.END

    async def start_conversation(self, update: Update, context: CallbackContext):
        # Answer callback
        if update:
            query = update.callback_query
            if query:
                await query.answer()

        # Check role
        if not self.restrict_function(update, context):
            return ConversationHandler.END

        keyboard = get_keyboard(context, back_button_callback="end") 
        await send_message(update, context, keyboard=keyboard, text_string_index=self.put_file_message)

        file_path = self.start_conversation_get_file_path(context)
        await self.file_handler.send_file_general(update, context, file_path)

        return self.INPUT_FILE

    async def file_error_response(
        self, update: Update, context: CallbackContext, file_path, text_string_index
    ):
        if not self.restrict_function(update, context):
            return ConversationHandler.END 

        await self.file_handler.delete_file(file_path)
        keyboard = get_keyboard(
            context,
            back_button_callback="end", 
        )
        await send_message(
            update, context, keyboard=keyboard, text_string_index=text_string_index
        )
        return self.INPUT_FILE

    async def receive_file(self, update: Update, context: CallbackContext):
        if not self.restrict_function(update, context):
            return ConversationHandler.END

        try:
            file_path = await self.file_handler.download_file(update)
            path = os.path.join("./data", file_path)
            result = self.file_handler.validate_file(path)
            code, message_text_id = (
                result[0],
                result[1],
            )  # there are more args sometimes
            if code != 200:
                return await self.file_error_response(
                    update, context, file_path, message_text_id
                )

            # Success
            await self.file_handler.add_path(context, file_path)
            self.file_handler.put_file_iterable(context)

            await end_conversation(
                    update,
                    context,
                    last_message_text_string_index=message_text_id,
                )
            return await self.end(update, context)

        except Exception as e:
            logger.error(f"Error recieving the file: {str(e)}")
            return await self.file_error_response(
                update, context, file_path, self.error_message
            )

    def register_handler(self):
        register_conversation_handler(self.bot_handler)


# My handlers
put_text_file_conv = PutFileConversation(text_handler, ADD_TEXT_BUTTON)
put_text_file_conv.register_handler()



import os

from .utils.init_templates import initialize_file_from_draft


# Get sectret
def read_env(file: bool):
    """Load environment variables from a .env file or from the system."""
    if file:
        env_vars = {}
        path = os.path.join(os.getcwd(), ".env")
        with open(path, "r") as file:
            for line in file:
                if line.strip() and not line.startswith("#"):
                    k, v = line.strip().split("=")
                    env_vars[k] = v
        return env_vars
    return dict(os.environ)


def get_secret_by_name(name: str, default = None, env_file = False) -> str:
    """Leave default = None to raise an error if the variable is not found."""
    if env_file:
        if not os.path.exists(os.path.join(os.getcwd(), ".env")):
            initialize_file_from_draft(".env", os.getcwd())
            raise Exception(
                "No .env file found in the current working directory. Please fill .env file."
            )
    vars = read_env(env_file)
    result = vars.get(name, default)
    if result is None:
        raise ValueError(f"Error initializing env variable: {name}")
    return result


def get_secret(name: str, default=None):
    """Leave default = None to raise an error if the variable is not found."""
    try:
        result = get_secret_by_name(name, default, env_file = False)
    except ValueError:
        result = get_secret_by_name(name, default, env_file = True)
    return result


# TG creds
BOT_NAME = get_secret("BOT_NAME")
BOT_TOKEN = get_secret("BOT_TOKEN")

# Optional
# File storage/mount
TIME_ZONE = get_secret("TIME_ZONE", "")

# Roles
DEFAULT_ROLES_DICT = {}

# Possible values: superadmin, admin, user, banned
def default_roles(roles):
    DEFAULT_ROLES_DICT.update({str(get_secret(k)): v for k, v in roles.items()})

def get_default_role(user_id):
    return DEFAULT_ROLES_DICT.get(str(user_id), "user")



from telegram import Update, InlineKeyboardButton, InlineKeyboardMarkup
from telegram.ext import CallbackContext, ConversationHandler
# from uuid import uuid1

from .context_logic import get_user_data, clear_cache
from ..text_handler import text_handler
from ..send import send_message
from ..mylogging import logger, get_time


# Info utils
ESSENTIAL_INFO_PREFIX = "tg"
def get_user_open_tg_info(user) -> dict:
    # Gather user info from telegram user object
    user_info = {
        # "uuid": uuid1(),  # additional anonimization
        f"{ESSENTIAL_INFO_PREFIX}_added_on": get_time(string=True),
        f"{ESSENTIAL_INFO_PREFIX}_username": f"@{user.username}",
        f"{ESSENTIAL_INFO_PREFIX}_first_name": user.first_name,
        f"{ESSENTIAL_INFO_PREFIX}_last_name": user.last_name,
    }
    return user_info


def get_user_hidden_tg_info(contact) -> dict:
    # Gather user info from telegram message.contact object when shared
    user_info = {
        f"{ESSENTIAL_INFO_PREFIX}_phone_number": contact.phone_number,
        "vcard": contact.vcard,
    }
    return user_info


def get_full_info(update):
    user = update.effective_user
    user_info = get_user_open_tg_info(user)
    contact = update.message.contact
    if contact:
        contact_info = get_user_hidden_tg_info(contact)
        user_info.update(contact_info)
    return {k: v for k, v in user_info.items() if v is not None}


def put_info_to_user_data(update, context):
    info_dict = get_full_info(update)
    user_data = get_user_data(context)
    user_data.update(info_dict)


def is_info_in_user_data(update, context):
    user_data = get_user_data(context)
    for k in user_data:
        if k.startswith(ESSENTIAL_INFO_PREFIX):
            return True


def get_user_data_essential(context, user_id):
    user_data = get_user_data(context, user_id)
    info_lines = [f"{k}: {v}" for k, v in user_data.items() if k.startswith(ESSENTIAL_INFO_PREFIX)]
    return "\n".join(info_lines)


# Keyboard utils
def get_keyboard(
    context,
    options=[],
    prefix="",
    back_button_callback="",
    back_button_text_string_index="back_button", # text.xlsx
    user_id=None,
):

    buttons = []

    if prefix:
        buttons += [
            [
                InlineKeyboardButton(
                    option,
                    callback_data=f"{prefix}_{option}",
                )
            ]
            for option in options
        ]

    if back_button_callback:
        buttons += [
            [
                InlineKeyboardButton(
                    text_handler.get_text(
                        context,
                        text_string_index=back_button_text_string_index,
                        user_id=user_id,
                    ),
                    callback_data=back_button_callback,
                )
            ]
        ]

    return InlineKeyboardMarkup(buttons) if buttons else None


async def get_info_from_query(update, prefix):
    if update:
        query = update.callback_query
        if query:
            # Get data
            query_data = query.data
            return query_data.replace(f"{prefix}_", "")
    logger.warning("get_info_from_query got no Update")


# Conversation utils
async def end_conversation(
    update: Update,
    context: CallbackContext,
    clean_up=False,
    last_message_text_string_index=None,
):

    if last_message_text_string_index is not None:
        await send_message(
            update, context, text_string_index=last_message_text_string_index, replace=False
        )

    if clean_up:
        clear_cache(context)

    return ConversationHandler.END


async def cancel(update: Update, context: CallbackContext):
    return await end_conversation(
        update, context, clean_up=True, last_message_text_string_index="cancel_response"
    )  # text.xlsx



import contextlib
import asyncio
import json
import os

from telegram.ext import Application, ApplicationBuilder, PicklePersistence
from telegram import Update

from fastapi import FastAPI, Request, Response
from http import HTTPStatus
import uvicorn

# Suppress uvicorn logger
import logging
uvicorn_logger = logging.getLogger("uvicorn")
uvicorn_logger.setLevel(logging.ERROR)

# Suppress pyngrok and fastapi loggers
pyngrok_logger = logging.getLogger("pyngrok")
pyngrok_logger.setLevel(logging.ERROR)

fastapi_logger = logging.getLogger("fastapi")
fastapi_logger.setLevel(logging.ERROR)

from .utils.init_templates import initialize_file_from_draft
from .mylogging import time_log_decorator, logger
from . import settings

# add handlers just with decorators throughout the code
from .decorators import add_handlers


# -------------------------------------------TELEGRAM BOT DB------------------------------
class MyApplication(Application):
    # Disagreeing with the user/chat_data read-only policy
    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.user_data = self._user_data
        self.chat_data = self._chat_data

    async def stop(self):
        # TODO handle urgent shutdown
        # Ensure all user data and chat data are saved
        if self.persistence:
            for user_id, data in self.user_data.items():
                await self.persistence.update_user_data(user_id, data)
            for chat_id, data in self.chat_data.items():
                # Sanitize for unpickable objects: convention
                for k in list(data.keys()):
                    if k.endswith("_func"):
                        data.pop(k)
                await self.persistence.update_chat_data(chat_id, data)

        # Call the original stop method
        await super().stop()


# --------------------------------------------TELEGRAM APP---------------------------------
def prepare_app():
    root_dir = "./data"
    initialize_file_from_draft("text.xlsx", root_dir)
    persistence = PicklePersistence(
        filepath=os.path.join(root_dir, f"{settings.BOT_NAME}Persistence"),
        on_flush=True,
    )
    application = (
        ApplicationBuilder()
        .application_class(MyApplication)
        .token(settings.BOT_TOKEN)
        .persistence(persistence)
        .build()
    )

    return application

# App instance
application = prepare_app()

# Operations
@time_log_decorator
def telegram_bot_polling(debug = False):
    # Add handlers; needs to be called at the end for the decorators reason
    add_handlers(application, debug)
    # Ignore exception when Ctrl-C is pressed
    with contextlib.suppress(KeyboardInterrupt):  
        application.run_polling()


# AWS
@time_log_decorator
async def process_update(event, context):
    # Add handlers; needs to be called at the end for the decorators reason
    add_handlers(application)
    try:
        await application.initialize() # .start() ?
        await application.process_update(
            Update.de_json(json.loads(event["body"]), application.bot)
        )
        await application.stop()
        return {"statusCode": 200, "body": "Success"}

    except Exception as e:
        logger.error(f"Error in process_update: {e}")
        return {"statusCode": 500, "body": "Failure"}


def lambda_handler(event, context):
    return asyncio.get_event_loop().run_until_complete(process_update(event, context))


# Vultr and others
def ptb_for_webhook():
    root_dir = "./data"
    persistence = PicklePersistence(
        filepath=os.path.join(root_dir, f"{settings.BOT_NAME}Persistence"),
        on_flush=True,
    )
    ptb = (
        ApplicationBuilder()
        .application_class(MyApplication)
        .updater(None)
        .token(settings.BOT_TOKEN)
        .persistence(persistence)
        .read_timeout(7)
        .get_updates_read_timeout(42)
        .build()
    )
    add_handlers(ptb)
    return ptb


def app_for_webhook(url, cert_file_location=None, secret_token=None):
    @contextlib.asynccontextmanager
    async def lifespan(app: FastAPI):
        if cert_file_location is None:
            await app.ptb.bot.setWebhook(url=url, secret_token=secret_token)
        else:
            with open(cert_file_location, "r") as cert_file:
                await app.ptb.bot.setWebhook(
                    url=url, certificate=cert_file, secret_token=secret_token
                )
        async with app.ptb as ptb:
            await ptb.start()
            yield
            await ptb.stop()

    # Initialize FastAPI app (similar to Flask)
    app = FastAPI(lifespan=lifespan)
    # TODO Limiter ? or nginx set up
    app.ptb = ptb_for_webhook()

    @app.post("/")
    async def process_update(request: Request):
        req = await request.json()
        update = Update.de_json(req, app.ptb.bot)
        try:
            await app.ptb.process_update(update)
            return Response(status_code=HTTPStatus.OK)
        except Exception as e:
            logger.error(f"Error in process_update: {e}")
            return Response(status_code=HTTPStatus.INTERNAL_SERVER_ERROR)

    return app

# Launch
@time_log_decorator
def run_webhook_uvicorn(access_log=True):
    """
    - requires ssl folder;

    - env:
        WEBHOOK_URL
        CERT_FILE_PATH
        LISTEN
        PORT
        SECRET_TOKEN (optional)

    - Nginx set up if there is a secret token:
        # nginx.conf
        location / {
            proxy_pass http://testbot:88;
            proxy_set_header X-Telegram-Bot-Api-Secret-Token your_secret_token;
        }

    - Check webhook status
        https://api.telegram.org/bot<your_bot_token>/getWebhookInfo
    """
    # secrets
    webhook_url = settings.get_secret("WEBHOOK_URL")
    cert_file_path = settings.get_secret("CERT_FILE_PATH", "") or None
    # optional
    listen = settings.get_secret("LISTEN", "0.0.0.0")  # we're in a docker container
    port = settings.get_secret("PORT", "80")
    secret_token = settings.get_secret("SECRET_TOKEN", "") or None

    # run
    app = app_for_webhook(webhook_url, cert_file_path, secret_token)
    uvicorn.run(app, host=listen, port=int(port), access_log=access_log)


# TODO big load solution
def run_webhook_gunicorn(env_file=False):
    """This might interfere with the convrsation handlers and spawn separate contexts."""
    webhook_url = settings.get_secret("WEBHOOK_URL", env_file=env_file)
    cert_file_path = settings.get_secret("CERT_FILE_PATH", env_file=env_file)
    listen = settings.get_secret(
        "LISTEN", env_file=env_file
    )  # we're in a docker container
    port = settings.get_secret("PORT", env_file=env_file)
    # optional
    secret_token = settings.get_secret("SECRET_TOKEN", env_file=env_file)  # or None

    global app
    app = app_for_webhook(webhook_url, cert_file_path, secret_token)

    file_name = __file__.split("/")[-1].split(".")[0]
    command = " ".join(
        [
            "gunicorn",
            "-w",
            "4",
            "-k",
            "uvicorn.workers.UvicornWorker",
            "-b",
            f"{listen}:{port}",
            f"{file_name}:app",
        ]
    )
    os.system(command)



import click
import os

from .mylogging import logger
from .utils.init_templates import initialize_file_from_draft, create_vultr_deploy_yml_from_draft
from . import settings


# Main
@click.group()
@click.version_option()
def cli():
    """The CLI tool."""


# Builds neccessary files for the developer and runs polling
@click.option("--docker", is_flag=True, help="Deploy as a Docker container; env vars from .env")
@cli.command()
def run(docker):
    root_dir = "./data"
    initialize_file_from_draft("text.xlsx", root_dir)

    root_dir = os.getcwd()
    initialize_file_from_draft("settings.py", root_dir)
    initialize_file_from_draft("main.py", root_dir)
    initialize_file_from_draft(".gitignore", root_dir)
    if docker:
        initialize_file_from_draft("Dockerfile", root_dir)
        initialize_file_from_draft(".dockerignore", root_dir)
        initialize_file_from_draft("docker-compose.yml", root_dir)
        run_command = f"docker-compose up"
        print(f"Executing: {run_command}")
        run_status = os.system(run_command)
        if run_status != 0:
            print("Failed to run Docker container.")
            return
        print("Container started successfully.")
        return
    # poetry ads
    try:
        os.system("poetry run python main.py")
    except Exception as e:
        logger.error(f"Please install poetry, or launch main.py directly yourself. Sorry! Error:\n{e}")


# Get the webhook info for a Telegram Bot
@click.option("--token", default=settings.BOT_TOKEN)
@cli.command()
def get_webhook_info(token):
    os.system(f"curl -X POST https://api.telegram.org/bot{token}/getWebhookInfo")


# Generate a secret token for Telegram
@cli.command()
def generate_secret_token():
    import secrets
    token = secrets.token_hex(16)
    logger.info(f"Generated secret token: {token}")


@cli.command()
def vultr():
    """Deploys to Vultr using Actions."""
    root_dir = os.getcwd()
    initialize_file_from_draft("Dockerfile", root_dir)
    initialize_file_from_draft(".dockerignore", root_dir)
    initialize_file_from_draft("docker-compose.yml", root_dir)
    create_vultr_deploy_yml_from_draft()


# Developer mode
if __name__ == "__main__":
    cli()



from telegram.ext import (
    CommandHandler,
    CallbackQueryHandler,
    ConversationHandler,
    MessageHandler,
    filters,
)
from functools import wraps
from uuid import uuid1

from .roles import check_role, DEFAULT_ALLOWED_ROLES


COMMAND_HANDLERS = {}
CALLBACK_HANDLERS = {}
CONVERSATION_HANDLERS = {}
MESSAGE_HANDLERS = {}


# Decorators
def command(name=None, allowed_roles = DEFAULT_ALLOWED_ROLES):
    def decorator(func):
        command_name = name or func.__name__

        @wraps(func)
        async def wrapper(update, context, *args, **kwargs):
            if not check_role(context, allowed_roles):
                return ConversationHandler.END
            return await func(update, context, *args, **kwargs)

        COMMAND_HANDLERS[command_name] = wrapper
        return wrapper
    return decorator


def button_callback(prefix=None, allowed_roles = DEFAULT_ALLOWED_ROLES):
    def decorator(func):
        callback_name = prefix or func.__name__
        callback_name = rf"^{callback_name}"

        @wraps(func)
        async def wrapper(update, context, *args, **kwargs):
            # Answer query
            if update:
                query = update.callback_query
                if query:
                    await query.answer()

            if not check_role(context, allowed_roles):
                return ConversationHandler.END

            return await func(update, context, *args, **kwargs)

        CALLBACK_HANDLERS[callback_name] = wrapper
        return wrapper
    return decorator


def message_handler(prefix=None, allowed_roles = DEFAULT_ALLOWED_ROLES):
    """One per app, for now"""
    def decorator(func):
        # TODO prefix
        name = "message_handler"

        @wraps(func)
        async def wrapper(update, context, *args, **kwargs):
            if update:
                if not update.message:
                    return False
                else:
                    if not update.message.text.strip():
                        return False
            else:
                return False

            if not check_role(context, allowed_roles):
                return ConversationHandler.END
            return await func(update, context, *args, **kwargs)

        MESSAGE_HANDLERS[name] = wrapper
        return wrapper
    return decorator


# Utils
def register_conversation_handler(handler):
    # required for persistence
    name = handler.name or str(uuid1())
    CONVERSATION_HANDLERS[name] = handler


def add_handlers(application, debug=False):
    # convos have to go first
    for _, h in CONVERSATION_HANDLERS.items():
        application.add_handler(h)

    for _, h in MESSAGE_HANDLERS.items():
        application.add_handler(
            MessageHandler(filters.TEXT & ~filters.COMMAND, h)
        )

    for k, v in COMMAND_HANDLERS.items():
        application.add_handler(CommandHandler(k, v))

    for k, v in CALLBACK_HANDLERS.items():
        application.add_handler(CallbackQueryHandler(v, k))

    if debug:
        info_lines = ["HANDLERS REGISTERED"]
        names = ["conversations", "messages", "commands", "button callbacks"]
        dicts = [
            CONVERSATION_HANDLERS,
            MESSAGE_HANDLERS,
            COMMAND_HANDLERS,
            CALLBACK_HANDLERS,
        ]
        for name, handler_dict in zip(names, dicts):
            if handler_dict:
                info_lines.append(f"{name}:")
                for k, v in handler_dict.items():
                    info_lines.append(f"  {k}: {v}")
        print("\n".join(info_lines))  # avoid logging module here to keep logs clean



from telegram import Update, InlineKeyboardMarkup, InlineKeyboardButton
from telegram.ext import CallbackContext

from .roles import DEFAULT_ADMIN_ROLES
from .decorators import command

from .text_handler import text_handler
from .send import send_message
from .utils.utils import clear_cache


# List of buttons; must be unique and present in the text.xlsx;
# callbacks are the same as the button names
ADMIN_MAIN_MENU_HEADER = "admin_menu_header"
ADD_TEXT_BUTTON = "add_text_button"
ADD_INTRO_VID_BUTTON = "add_intro_video_button"
MANAGE_USERS_BUTTON = "admin_manage_users_button"
MAILING_BUTTON = "mailing_users_button"
ADMIN_MENU = [
    ADD_TEXT_BUTTON,
    ADD_INTRO_VID_BUTTON,
    MANAGE_USERS_BUTTON,
    MAILING_BUTTON,
]


@command(allowed_roles=DEFAULT_ADMIN_ROLES)
async def admin(update: Update, context: CallbackContext):
    # each represent a conversation handler, for example add_text_conv.py,
    # which is then connected at main.py
    clear_cache(context) # refresh cache
    options = {
        text_handler.get_text(context, option): option for option in ADMIN_MENU
    }

    buttons = [
        [
            InlineKeyboardButton(
                option_name,
                callback_data=f"{option}",
            )
        ]
        for option_name, option in options.items()
    ]
    keyboard = InlineKeyboardMarkup(buttons)
    return await send_message(update, context, keyboard=keyboard, text_string_index=ADMIN_MAIN_MENU_HEADER, new=True)



import easy_tg_bot

# initiate other modules
import settings


@easy_tg_bot.command()
async def help(update, context):
    return await easy_tg_bot.send_message(update, context, text_string_index = "help_message", replace=False)  # text.xlsx


if __name__=="__main__":
    easy_tg_bot.start.START_DONE_CALLBACK = help
    easy_tg_bot.telegram_bot_polling()
    # easy_tg_bot.run_webhook_uvicorn()



from telegram.ext import (
    CallbackContext,
)

from .utils.context_logic import put_user_data, get_user_data
from .file_handler import FileHandler
from .validate_text_file import (
    get_text_dictionary,
    get_text_from_dict,
    get_text_and_parse_mode_from_dict,
    get_languages_from_dict,
    get_main_language_from_dict,
    validate_text_file,
)


# Const
TEXT_PREFIX = "text"


class TextFileHandler(FileHandler):
    def __init__(self):
        super().__init__(TEXT_PREFIX, get_text_dictionary, validate_text_file)

    # Language logic
    def get_languages(self, context: CallbackContext):
        return get_languages_from_dict(self.get_file_iterable(context))

    def get_main_language(self, context: CallbackContext):
        return get_main_language_from_dict(self.get_file_iterable(context))

    # User Language logic
    def assign_language_to_user(self, context, lan):
        put_user_data(context, "lan", lan)

    def get_user_language(self, context, user_id=None):
        user_data = get_user_data(context, user_id)
        return user_data.get("lan", self.get_main_language(context))

    # Get Text
    def get_text(self, context: CallbackContext, text_string_index: str, user_id=None):
        language = self.get_user_language(context, user_id)
        dic = self.get_file_iterable(context)
        return get_text_from_dict(language, dic, text_string_index)

    def get_text_and_parse_mode(self, context, text_string_index: str, user_id=None):
        language = self.get_user_language(context, user_id)
        dic = self.get_file_iterable(context)
        return get_text_and_parse_mode_from_dict(language, dic, text_string_index)


text_handler = TextFileHandler()
